// Meccano MAX HiCAT Livera Demo
#include <Servo.h>
#include <MeccaMaxServoDevice.h>
#include <MeccaMaxFaceDevice.h>
#include <MeccaMaxDrive.h>
#include <Scheduler.h>
#include <SerialCommand.h>

#define _MOTORS
#define _FACE
//#define _SERVOS

static byte happyFace[18] = {
  0xFF, 0xFF, 0xF1, 0xE6, 0xE4, 0xB1, 0x7F, 0x7F,
  0x7F, 0x7F, 0xF1, 0xE6, 0xE4, 0xF1, 0xFF, 0xFF,
  0xF5, 0x2D
  };
static byte sadFace[18] = {
  0xFF, 0xFF, 0xFF, 0xC7, 0x8B, 0x8D, 0xC3, 0xFF, 
  0xFF, 0xC3, 0x8D, 0x8B, 0xC7, 0xFF, 0xFF, 0xFF, 
  0xF5, 0x03 
};

// for use by Face FSM for multitasking purposes
enum FaceStates {
  SAD_FACE,
  HAPPY_FACE
};

// use up all PWM pins because they have connector built-in
int pinFACE = MISO;  // unused MISO pin
int pinSERVOS = 12;  // unused pin 12
int pinDRIVE = SCK;  // unused sck pin
int PWM1 = 13;       // PWM4
int PWM_CE = 11;     // servo enable

// use motor driver M2
int BIN3 = A2;  //control laser
int BIN4 = A3;  //control laser

int nextUpdate = 500;
static int pos = 0;

bool motorsReady = false;
static FaceStates faceState = SAD_FACE;

Servo *myservo;
SerialCommand sCmd;

MeccaChannel *channelFACE;
MeccaChannel *channelSERVOS;
MeccaChannel *channelDRIVE;
MeccaMaxServoDevice *headServo;
MeccaMaxServoDevice *clawServo;
MeccaMaxFaceDevice *face;
MeccaMaxDrive *drive;
Scheduler motorScheduler = Scheduler();
Scheduler mainScheduler = Scheduler();

uint8_t cameraAngle = 120;
uint8_t cameraAngleRange[] = {
  60, 150
};
uint8_t laserIsOn = 0;

// print protocol exchange between channel and devices
void printDebugInfo(MeccaChannel *channel)
{
  for (int i = 0; i < MeccaChannel :: MAX_DEVICES_PER_CHANNEL; i++) 
  {
    Serial.print(channel->getOutputByteInfo(i), HEX);
    Serial.print(" ");
  }
  Serial.print(channel->getModuleNumInfo(), HEX);
  Serial.print(" - ");
  Serial.println(channel->getInputByteInfo(), HEX);
}
void setup() {

  
  Serial.begin(115200);
  Serial1.begin(115200);

  Serial.println("-----Meccano Max at your service----");

  
  // enable servo block
  pinMode(PWM_CE, OUTPUT);
  digitalWrite(PWM_CE, HIGH);  

  myservo = new Servo();
  myservo->attach(PWM1);
  myservo->write(cameraAngle);
  // disable servo block so as not to affect servos while initializing below
  // otherwise the connected servo will be sent spurious signals
  // remember to enable it later or servo will not work
  //digitalWrite(PWM_CE, LOW);
  
  // initialize the lazer.
  pinMode(BIN3, OUTPUT);
  pinMode(BIN4, OUTPUT);
  digitalWrite(BIN3, LOW);  // turn off LASER
  digitalWrite(BIN4, LOW);

  // Setup callbacks for SerialCommand commands
  sCmd.addCommand("laser", toggleLaser);
  sCmd.addCommand("camera", positionCamera);
  sCmd.addCommand("motor", driveMotors);
  sCmd.setDefaultHandler(printError);
#ifdef _MOTORS
  channelDRIVE = new MeccaChannel(pinDRIVE);
  drive = new MeccaMaxDrive(channelDRIVE);
    // wait till last device in chain has been discovered
  while (channelDRIVE->getDeviceName(1) == NULL) {
    channelDRIVE->communicate(); // start comm with devices
    printDebugInfo(channelDRIVE);
    //sCmd.readSerial((Stream&)Serial1);
    delay(100);
  }
  motorsReady = true;
#endif

#ifdef _SERVOS
  // setup 2 servos
  channelSERVOS = new MeccaChannel(pinSERVOS);
  headServo = new MeccaMaxServoDevice("HeadServo");
  headServo->mapTo180Degrees(100, 150);
  clawServo = new MeccaMaxServoDevice("ClawServo");
  clawServo->mapTo180Degrees(0x90, 0xe0);
  // using Chain of Responsibility Design Pattern to loosely couple Channel and Devices
  headServo->setSuccessor(clawServo);
  channelSERVOS->setDeviceChain(headServo);
  while (channelSERVOS->getDeviceName(1) == NULL) {
    channelSERVOS->communicate(); // start comm with devices
    //sCmd.readSerial((Stream&)Serial1);
    delay(100);
  }
  mainScheduler.schedule(testServos, 200);
#endif

#ifdef _FACE
// set up 1 face device
  channelFACE = new MeccaChannel(pinFACE);
  face = new MeccaMaxFaceDevice("Face");
  //face->setScrollImage(scrollMaxImage, 70);
  channelFACE->setDeviceChain(face);

  //printImage(happyFace, 18);
  //printImage(sadFace, 18);
  happyFace[17] = channelFACE->calculateImageCheckSum(happyFace, 17);
  sadFace[17] = channelFACE->calculateImageCheckSum(sadFace, 17);
  while (channelFACE->getDeviceName(0) == NULL) {
    channelFACE->communicate(); // start comm with devices
    //sCmd.readSerial((Stream&)Serial1);
    delay(100);
  }
  channelFACE->setMaxDevices(1);
  mainScheduler.schedule(testFace, 200);
#endif

  // note that it is possible to chain 2 motors and 2 servos in one channel
  // max. 4 devices per channel
  

  Serial.println("Completed setup...");
}

// the loop routine runs over and over again forever:
void loop() {
  sCmd.readSerial((Stream&)Serial1);
  mainScheduler.update();
  motorScheduler.update();
}

// ****Add Face and servo processing function here****
#ifdef _FACE
void testFace() 
{

    // FSM sequencing the display of scrolling text and the happy face
    switch (faceState) {
      case SAD_FACE:
        // display sad face
        face->setImage(sadFace, 18, channelFACE);
        mainScheduler.schedule(testFace, 2000);
        faceState = HAPPY_FACE;
        break;
      case HAPPY_FACE:
        // display happy face
        face->setImage(happyFace, 18, channelFACE);
        mainScheduler.schedule(testFace, 3000);
        faceState = SAD_FACE;
        break;
    }
}
#endif

#ifdef _SERVOS
void testServos() 
{
  Serial.print("Servo position: ");
  Serial.println(pos);

  // make the servos move 30 degrees at a time between 0 and 180
  //printDeviceStatus(headServo, clawServo);
  //printDevices(channelSERVOS);
  headServo->setServoPosition(channelSERVOS, pos);
  clawServo->setServoPosition(channelSERVOS, pos);
  channelSERVOS->communicateAll();
  pos = (pos + 30) % 180;
  mainScheduler.schedule(testServos, 1500);
}
#endif

// ****End of Face and servo processing function****

// command handler - tiggle laser on and off
void toggleLaser()
{
  if (laserIsOn) {
    digitalWrite(BIN3, LOW);  // turn off LASER
    laserIsOn = 0;
    Serial.println("laser off");
  }
  else {
    digitalWrite(BIN3, HIGH);  // turn on LASER
    laserIsOn = 1;
    Serial.println("laser on");
  }
}

// command handler - tilt camera up and down
void positionCamera()
{
  char *param;

  param = sCmd.next();
  if (!strcmp(param, "down")) {
    if (cameraAngle < cameraAngleRange[1]) {
      cameraAngle += 10;

      //digitalWrite(PWM_CE, HIGH);
      myservo->write(cameraAngle);
      //digitalWrite(PWM_CE, LOW);
    }
  }
  else if (!strcmp(param, "up")) {
    if (cameraAngle > cameraAngleRange[0]) {
      cameraAngle -= 10;

      //digitalWrite(PWM_CE, HIGH);
      myservo->write(cameraAngle);
      //digitalWrite(PWM_CE, LOW);
    }
  }

  Serial.print("camera angle: ");
  Serial.println(cameraAngle);
}

// command handler - drive motors according to left/right values
void driveMotors()
{
  char *param;
  int16_t left = 0;
  int16_t right = 0;

  param = sCmd.next();
  if (param == NULL) {
    return;
  }
  left = atoi(param);

  param = sCmd.next();
  if (param == NULL) {
    return;
  }
  right = atoi(param);

  Serial.print("driveMotors received: motor ");
  Serial.print(left);
  Serial.print(" ");
  Serial.println(right);
  int state = 0;

  
  // stop all scheduled motor update
  motorScheduler.clear();

  //if (!motorsReady) return;
  Serial.println("setting motor direction");

  // Hicat webapp always send "0 0" after every movement command
  // need to ignore it for now
  //if ((left == 0) && (right == 0)) return;
  //driveForward();
  //return;
  
  // adapt to the way the existing HiCAT application works
  if (left > 30) {
    // forward
    //analogWrite(AIN2, left);
    //digitalWrite(AIN1, LOW);
    state += 1;
  } else if (left < -30)  {
    // reverse
    state += 2;
    //analogWrite(AIN2, 255 + left);
    //digitalWrite(AIN1, HIGH);
  }
  //else {
    // motor off
    //digitalWrite(AIN1, LOW);
    //digitalWrite(AIN2, LOW);
  //}

  if (right > 30) {
    // forward
    state += 4;
    //analogWrite(AIN3, right);
    //digitalWrite(AIN4, LOW);
  }
  else if (right < -30) {
    // reverse
    state += 8;
    //analogWrite(AIN3, 255 + right);
    //digitalWrite(AIN4, HIGH);
  }
  //else {
    // motor off
    //digitalWrite(AIN3, LOW);
    //digitalWrite(AIN4, LOW);
  //}
  switch (state) {
    case 0:
      //driveStop();
      motorScheduler.schedule(driveStop, nextUpdate);
      break;
    case 5:
      //driveForward();
      drive->forward();
      break;
    case 2:
    case 4:
    case 6:
      //driveLeft();
      drive->turnLeft();
      break;
    case 1:
    case 8:
    case 9:
      //driveRight();
      drive->turnRight();
      break;
    case 10:
      //driveBackward();
      drive->backward();
      break;
  }
}

// command handler - error: not matching any commands
void printError(const char *command) {
  Serial.print("Invalid command: ");
  Serial.print(command);
  Serial.println('?');
}

// function called by driveMotors
// we need a scheduler to execute these commands because the MAX
// motor will stop if the command is not executed periodically
/*
// the functions should be self-explanatory
void driveForward()
{
  drive->forward();
  //motorScheduler.schedule(driveForward, nextUpdate);
}
void driveLeft()
{
  drive->turnLeft();
  //motorScheduler.schedule(driveLeft, nextUpdate);
}

void driveRight()
{
  drive->turnRight();
  //motorScheduler.schedule(driveRight, nextUpdate);
}
void driveBackward()
{
  drive->backward();
  //motorScheduler.schedule(driveBackward, nextUpdate);
}
*/
void driveStop()
{
  drive->stop();
}


